////////////////////////////////////////////////////////////////////////////////
///
///		Title		: NPhoto Fixed size Memory Management
///
///		Copyright (c) 2006, NHN Co., Ltd. ALL RIGHTS RESERVED
///
///		Team		: Desktop Service Development
///
////////////////////////////////////////////////////////////////////////////////
#pragma once

#include <vector>
static const int __DEFAULT_CHUNK_SIZE = 4096;
static const int __MAX_SMALL_OBJECT_SIZE = 64;

/**
* @group 	PCM Base Lib	
* @brief 	Fixed block size Memory management
* @date		08/21/2006
* @author	telltale
* @remark	Example\n
class CNPThumbNailBuffer\n
{\n
	BYTE*		Malloc(DWORD dwLen)\n
	{\n
		if (0 < dwLen && dwLen < 1024)\n
		{\n
			return (BYTE*)m_mem1KAllocator.Alloc();\n
		}\n
		else if (1024 <= dwLen && dwLen < 10240)\n
		{\n
			return (BYTE*)m_mem10KAllocator.Alloc();\n
		}\n
		else \n
		if (dwLen > 0)\n
		{\n
			return new BYTE[dwLen];\n
		}\n
		return NULL;\n
	}\n
	void		Mfree(BYTE* pMem, DWORD dwLen)\n
	{\n
		if (0 < dwLen && dwLen < 1024)\n
		{\n
			m_mem1KAllocator.Free(pMem);\n
		}\n
		else if (1024 <= dwLen && dwLen < 10240)\n
		{\n
			m_mem10KAllocator.Free(pMem);\n
		}\n
		else\n
		if (dwLen > 0)\n
		{\n
			delete[] pMem;\n
		}\n
	}\n
	private:	\n
	CNPMemoryPool<BYTE[1024]> m_mem1KAllocator;\n
	CNPMemoryPool<BYTE[10240]> m_mem10KAllocator;\n
	public:\n
	CNPThumbNailBuffer(void){}\n
	~CNPThumbNailBuffer(void){}\n
};\n
*/
template < class USER_TYPE>
class CNPMemoryPool
{
	struct Chunk
	{
		void Init(size_t blockSize, unsigned char blocks)
		{
			pData_ = new unsigned char[blockSize * blocks];
			Reset(blockSize, blocks);
		}

		void* Allocate(size_t blockSize)
		{
			if (!blocksAvailable_) return 0;


			unsigned char* pResult =
				pData_ + (firstAvailableBlock_ * blockSize);
			firstAvailableBlock_ = *pResult;
			--blocksAvailable_;

			return pResult;

		}

		void Deallocate(void* p, size_t blockSize)
		{
			unsigned char* toRelease = static_cast<unsigned char*>(p);
			// Alignment check

			*toRelease = firstAvailableBlock_;
			firstAvailableBlock_ = static_cast<unsigned char>(
				(toRelease - pData_) / blockSize);
			// Truncation check

			++blocksAvailable_;
		}

		void Reset(size_t blockSize, unsigned char blocks)
		{
			firstAvailableBlock_ = 0;
			blocksAvailable_ = blocks;

			unsigned char i = 0;
			unsigned char* p = pData_;
			for (; i != blocks; p += blockSize)
			{
				*p = ++i;
			}
		}

		void Release()
		{
			delete[] pData_;
		}

		unsigned char* pData_;
		unsigned char firstAvailableBlock_, blocksAvailable_;
	};

public:
	/**
	* @brief 	CNPMemoryPool constructor
	* @param	blockSize	memoryblock size
	* @param	nDefaultChunkSize	memorychunk size
	*/
	CNPMemoryPool(size_t blockSize = 0, size_t nDefaultChunkSize=__DEFAULT_CHUNK_SIZE)
	{
		m_BlockSize		= sizeof(USER_TYPE);//blockSize;
		m_AllocChunk		= 0;
		m_DeAllocChunk	= 0;

		size_t numBlocks = nDefaultChunkSize / m_BlockSize;
		if (numBlocks > UCHAR_MAX) 
		{
			numBlocks = UCHAR_MAX;
		}
		else if (numBlocks == 0) 
		{
			numBlocks = 64;//8 * m_BlockSize;
		}

		m_NumBlocks = static_cast<unsigned char>(numBlocks);
	}

	~CNPMemoryPool()
	{
		Chunks::iterator i = m_Chunks.begin();
		for (; i != m_Chunks.end(); ++i)
		{
			i->Release();
		}
	}
	/**
	* @brief 	get fixed size memory from memorypool
	* @return	fixed size memory
	*/
	void* alloc()
	{
		Alloc();
	}
	void* Alloc()
	{
		if (m_AllocChunk == 0 || m_AllocChunk->blocksAvailable_ == 0)
		{
			Chunks::iterator i = m_Chunks.begin();
			for (;; ++i)
			{
				if (i == m_Chunks.end())
				{
					// Initialize
					m_Chunks.reserve(m_Chunks.size() + 1);
					Chunk newChunk;
					newChunk.Init(m_BlockSize, m_NumBlocks);
					m_Chunks.push_back(newChunk);
					m_AllocChunk = &m_Chunks.back();
					m_DeAllocChunk = &m_Chunks.front();
					break;
				}
				if (i->blocksAvailable_ > 0)
				{
					m_AllocChunk = &*i;
					break;
				}
			}
		}

		return m_AllocChunk->Allocate(m_BlockSize);
	}
	/**
	* @brief 	release memory from fixed size memorypool
	* @param	memory from fixed size memorypool
	*/
	void free(void* p)
	{
		Free(p);
	}
	void Free(void* p)
	{
		m_DeAllocChunk  = VicinityFind(p);

		DoDeallocate(p);
	}

	/**
	* @brief 	clear memorypool
	*/
	void release()
	{
		Release();
	}
	void Release()
	{
		Chunks::iterator i = m_Chunks.begin();
		for (; i != m_Chunks.end(); ++i)
		{
			i->Release();
		}

		m_Chunks.clear();
	}

private:
	Chunk* vicinity_find(void* p)
	{
		return VicinityFind(p);
	}
	Chunk* VicinityFind(void* p)
	{
		const size_t chunkLength = m_NumBlocks * m_BlockSize;

		Chunk* lo = m_DeAllocChunk;
		Chunk* hi = m_DeAllocChunk + 1;
		Chunk* loBound = &m_Chunks.front();
		Chunk* hiBound = &m_Chunks.back() + 1;

		if (hi == hiBound) hi = 0;

		for (;;)
		{
			if (lo)
			{
				if (p >= lo->pData_ && p < lo->pData_ + chunkLength)
				{
					return lo;
				}
				if (lo == loBound) lo = 0;
				else --lo;
			}

			if (hi)
			{
				if (p >= hi->pData_ && p < hi->pData_ + chunkLength)
				{
					return hi;
				}
				if (++hi == hiBound) hi = 0;
			}
		}
		return 0;		
	}


	void DoDeallocate(void* p)
	{

		m_DeAllocChunk->Deallocate(p, m_BlockSize);

		if (m_DeAllocChunk->blocksAvailable_ == m_NumBlocks)
		{

			Chunk& lastChunk = m_Chunks.back();

			if (&lastChunk == m_DeAllocChunk)
			{
				//  ΰ Chunks ִ Ȯ.  ΰ  release Ѵ

				if (m_Chunks.size() > 1 && 
					m_DeAllocChunk[-1].blocksAvailable_ == m_NumBlocks)
				{
					// Two free chunks, discard the last one
					lastChunk.Release();
					m_Chunks.pop_back();
					m_AllocChunk = m_DeAllocChunk = &m_Chunks.front();
				}
				return;
			}

			if (lastChunk.blocksAvailable_ == m_NumBlocks)
			{
				// ΰ  , ϳ release
				lastChunk.Release();
				m_Chunks.pop_back();
				m_AllocChunk = m_DeAllocChunk;
			}
			else
			{
				// ִ Chunk ǵڷ ̵
				std::swap(*m_DeAllocChunk, lastChunk);
				m_AllocChunk = &m_Chunks.back();
			}
		}		
	}

	size_t BlockSize() const
	{ 
		return m_BlockSize; 
	}

	size_t m_BlockSize;
	unsigned char m_NumBlocks;
	typedef std::vector<Chunk> Chunks;
	Chunks m_Chunks;
	Chunk* m_AllocChunk;
	Chunk* m_DeAllocChunk;
};
